#include "yeti.h"

#define CACHE_ENABLE    asm("mov 2,r1 \n	ldsr r1,sr24": /* No Output */: /* No Input */: "r1" /* Reg r1 Used */)

#define CACHE_DISABLE    asm("ldsr r0,sr24")

/******************************************************************************/

void CODE_IN_IWRAM matrix_rotate_world(matrix_t m, int alp, int bet, int gam)
{
  int cosalp = fixcos(alp);
  int sinalp = fixsin(alp);
  int cosbet = fixcos(bet);
  int sinbet = fixsin(bet);
  int cosgam = fixcos(gam);
  int singam = fixsin(gam);

  m[0][0] =  fixmul(singam, fixmul(sinbet, sinalp)) + fixmul(cosgam, cosalp);
  m[0][1] =  fixmul(cosbet, sinalp);
  m[0][2] =  fixmul(singam, cosalp) - fixmul(cosgam, fixmul(sinbet, sinalp));

  m[1][0] =  fixmul(singam, fixmul(sinbet, cosalp)) - fixmul(cosgam, sinalp);
  m[1][1] =  fixmul(cosbet, cosalp);
  m[1][2] = -fixmul(cosgam, fixmul(sinbet, cosalp)) - fixmul(singam, sinalp);

  m[2][0] = -fixmul(singam, cosbet);
  m[2][1] =  sinbet;
  m[2][2] =  fixmul(cosgam, cosbet);
}

void CODE_IN_IWRAM matrix_rotate_object(matrix_t m, int alp, int bet, int gam)
{
  int cosalp = fixcos(alp);
  int sinalp = fixsin(alp);
  int cosbet = fixcos(bet);
  int sinbet = fixsin(bet);
  int cosgam = fixcos(gam);
  int singam = fixsin(gam);

  m[0][0] =  fixmul(cosalp, cosgam) - fixmul(sinalp, fixmul(sinbet, singam));
  m[0][1] =  fixmul(sinalp, cosgam) + fixmul(cosalp, fixmul(sinbet, singam));
  m[0][2] =  fixmul(cosbet, singam);

  m[1][0] = -fixmul(sinalp, cosbet);
  m[1][1] =  fixmul(cosalp, cosbet);
  m[1][2] = -sinbet;

  m[2][0] = -fixmul(cosalp, singam) - fixmul(sinalp, fixmul(sinbet, cosgam));
  m[2][1] =  fixmul(cosalp, fixmul(sinbet, cosgam)) - fixmul(sinalp, singam);
  m[2][2] =  fixmul(cosbet, cosgam);
}

/******************************************************************************/

/*
** Name: vertex_project
** Desc: Projects a vertex into screen space. (ie: 3D to 2D viewport space)
*/
void CODE_IN_IWRAM vertex_project(vertex_t* v)
{
  if (v->z > YETI_RECIPROCAL_MIN && v->z < YETI_RECIPROCAL_MAX)
  {
    int z = reciprocal[v->z];

    v->sx =  (((YETI_VIEWPORT_X2-YETI_VIEWPORT_X1)*v->x*z)>>9)+i2fdiv2(YETI_VIEWPORT_X2+YETI_VIEWPORT_X1);
    v->sy = -(((YETI_VIEWPORT_Y2-YETI_VIEWPORT_Y1)*v->y*z)>>9)+i2fdiv2(YETI_VIEWPORT_Y2+YETI_VIEWPORT_Y1);
  }
  else
  {
    v->sx = v->sy = 0;
  }
}

/******************************************************************************/

/*
** Name: draw_sprite
** Desc: Draws a 16bit sprite to a framebuffer. The sprite is clipped to the
**       given rectangle.
*/
#ifndef __PATCH_DRAW_SPRITE__
void CODE_IN_IWRAM draw_sprite(
  yeti_t* yeti,
  int x1, int y1, int x2, int y2,
  YETI_ROM sprite_t sprite, int u1, int v1, int u2, int v2,
  int cl, int ct, int cr, int cb, int mode)
{
  if ((x2 - x1) > 0 && (y2 - y1) > 0)
  {
    int uu = f2i((u2 - u1) * reciprocal[x2 - x1]);
    int vv = f2i((v2 - v1) * reciprocal[y2 - y1]);
    
    u1 = i2f(u1);
    v1 = i2f(v1);

    if (x1 < cl) {u1 += (cl - x1) * uu; x1 = cl;}
    if (y1 < ct) {v1 += (ct - y1) * vv; y1 = ct;}
    if (x2 > cr) {x2 = cr;}
    if (y2 > cb) {y2 = cb;}

    if ((x2 - x1) > 0 && (y2 - y1) > 0)
    {
      int w = sprite[0];

      switch (mode)
      {
        case 0:
        {
          for (; y1 < y2; y1++, v1 += vv)
          {
            YETI_ROM u16* tb = &sprite[f2i(v1)*w+2];
            //u16* fb = &yeti->viewport.back->pixels[y1][x1];
            int i = x2 - x1, u = u1 - uu;

//#define AFFINE(I) {u16 c=tb[f2i(u+=uu)];if(c)fb[I]=c;}
          //AFFINE_LOOP;
//#undef  AFFINE
          }
          break;
        }
        case 1:
        {
          for (; y1 < y2; y1++, v1 += vv)
          {
            YETI_ROM u16* tb = &sprite[f2i(v1)*w+2];
            //u16* fb = &yeti->viewport.back->pixels[y1][x1];
            int i = x2 - x1, u = u1 - uu;

//#define AFFINE(I) {u16 c=tb[f2i(u+=uu)];if(c)fb[I]=RGB_BLEND(c,fb[I]);}
          //AFFINE_LOOP;
//#undef  AFFINE
          }
          break;
        }
        case 2:
        {
          for (; y1 < y2; y1++, v1 += vv)
          {
            YETI_ROM u16* tb = &sprite[f2i(v1)*w+2];
            //u16* fb = &yeti->viewport.back->pixels[y1][x1];
            int i = x2 - x1, u = u1 - uu;

//#define AFFINE(I) {fb[I] = RGB_BLEND(tb[f2i(u+=uu)], fb[I]);}
          //AFFINE_LOOP;
//#undef  AFFINE
          }
          break;
        }
      }
    }
  }
}
#endif

/*
** Name: stretch_sprite_di
** Desc: Draws a device independant sprite to the screen. The sprite will be
**       scaled to match the viewport.
*/
void CODE_IN_IWRAM stretch_sprite_di(
  yeti_t* yeti,
  int x1, int y1, int x2, int y2, int z,
  YETI_ROM sprite_t sprite,
  int u1, int v1, int u2, int v2, int mode)
{
  if (z > 4)
  {
    vertex_t a, b;

    a.x = x1; a.y = y1; a.z = z;
    b.x = x2; b.y = y2; b.z = z;

    vertex_project(&a);
    vertex_project(&b);

    draw_sprite(
      yeti, f2i(a.sx), f2i(a.sy), f2i(b.sx), f2i(b.sy),
      sprite, u1, v1, u2, v2,
      YETI_VIEWPORT_X1, YETI_VIEWPORT_Y1, YETI_VIEWPORT_X2, YETI_VIEWPORT_Y2,
      mode);
  }
}

void CODE_IN_IWRAM draw_sprite_di(
  yeti_t* yeti,
  int x1, int y1, int x2, int y2, int z,
  YETI_ROM sprite_t sprite, int mode)
{
  stretch_sprite_di(yeti, x1, y1, x2, y2, z, sprite, 0, 0, sprite[0], sprite[1], mode);
}

/*
** Name: draw_poly
** Desc: Renders a lit, texture mapped polygon to a viewport. No clipping is applied. 
** Note: This is the function that Greg replaced with his poly ARM asm code.
*/
#ifndef __PATCH_DRAW_POLY__
void CODE_IN_IWRAM draw_poly(vertex_t* p, int n, int texture, u32* const dst)
{
  int lt_i, lt_y, lt_yy;
  int rt_i, rt_y, rt_yy;
  int y1, y2, i, offset, cofs;
  u32 ofsetchunk;
  int lt_length = 0, rt_length = 0;
  int x1 = 999999, x2 = -999999;

  CACHE_ENABLE;

  for (i = n; i--;)
  {
    p[i].d = fixup(p[i].x);
    if (p[i].x < x1) {x1 = p[i].x; lt_i = rt_i = i;}
    if (p[i].x > x2) {x2 = p[i].x;}
  }
  CACHE_DISABLE;

  if ((x1 = fixup(x1)) < YETI_VIEWPORT_X1) x1 = YETI_VIEWPORT_X1;
  if ((x2 = fixup(x2)) > YETI_VIEWPORT_X2) x2 = YETI_VIEWPORT_X2;

  CACHE_ENABLE;
  for (; x1 < x2; x1++)
  {
    if (--lt_length <= 0)
    {
      do {
        i = lt_i++; if (lt_i >= n) lt_i = 0;
      } while ((lt_length = p[lt_i].d - p[i].d) <= 0);

      lt_yy = r2i((p[lt_i].y - (lt_y = p[i].y)) * reciprocal[lt_length]);
    }
    if (--rt_length <= 0)
    {
      do {
        i = rt_i--; if (rt_i < 0) rt_i = n - 1;
      } while ((rt_length = p[rt_i].d - p[i].d) <= 0);

      rt_yy = r2i((p[rt_i].y - (rt_y = p[i].y)) * reciprocal[rt_length]);
    }
    if ((y1 = fixup(lt_y)) < YETI_VIEWPORT_Y1) y1 = YETI_VIEWPORT_Y1;
    if ((y2 = fixup(rt_y)) > YETI_VIEWPORT_Y2) y2 = YETI_VIEWPORT_Y2;

    if ((i = y2 - y1) > 0)
    {
//-----------------------------------------------------------------
//bitstring renderer

	offset = (x1 << 4) + (y1 >> 4);
	ofsetchunk = (y1 & 0x0000000f) << 1;
	cofs = ((x1 ^ y1) & 1) ? 2 : 0;

	asm("
		st.w	r29,-4[r3]
		mov	%0,r26
		mov	%1,r27
		mov	%2,r28
		mov	%3,r29
		mov	%4,r30
		.hword 0x7C0B
		ld.w	-4[r3],r29
	"
	:  /* Output */
	: "r" (ofsetchunk), "r" (cofs), "r" (i << 1), "r" (&dst[offset]), "r" (&britered[texture][0]) /* Input */
	: "r26", "r27", "r28", "r29", "r30" /* Clobber */
	);
//-------------------------------------------------------------------------
    }
    lt_y += lt_yy;
    rt_y += rt_yy;
  }

  CACHE_DISABLE;

}
#endif

void CODE_IN_IWRAM set_pixel(int x, int y, int color, u32* const dst)
{
  int offset = (x << 4) + (y >> 4);
  int bitpos = (y & 0x0f) << 1;
  dst[offset] &= ~(0x03 << bitpos); //Clear pixel
  dst[offset] |= (color << bitpos); //Set new color
}

void draw_line(int x1, int y1, int x2, int y2, int color, u32* const dst)
{
  int dy, dx;
  int stepx, stepy;

  if ((x1 = fixup(x1)) < YETI_VIEWPORT_X1) x1 = YETI_VIEWPORT_X1;
  if (x1 >= YETI_VIEWPORT_X2) x1 = YETI_VIEWPORT_X2-1;
  if ((x2 = fixup(x2)) < YETI_VIEWPORT_X1) x2 = YETI_VIEWPORT_X1;
  if (x2 >= YETI_VIEWPORT_X2) x2 = YETI_VIEWPORT_X2-1;
  if ((y1 = fixup(y1)) < YETI_VIEWPORT_Y1) y1 = YETI_VIEWPORT_Y1;
  if (y1 >= YETI_VIEWPORT_Y2) y1 = YETI_VIEWPORT_Y2-1;
  if ((y2 = fixup(y2)) < YETI_VIEWPORT_Y1) y2 = YETI_VIEWPORT_Y1;
  if (y2 >= YETI_VIEWPORT_Y2) y2 = YETI_VIEWPORT_Y2-1;

  dy = y2 - y1;
  dx = x2 - x1;

  if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
  if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
  dy <<= 1;
  dx <<= 1;

  set_pixel(x1, y1, color, dst);
  if (dx > dy) {
    int fraction = dy - (dx >> 1);
    while (x1 != x2) {
      if (fraction >= 0) {
        y1 += stepy;
        fraction -= dx;
      }
      x1 += stepx;
      fraction += dy;
      set_pixel(x1, y1, color, dst);
    }
  } else {
    int fraction = dx - (dy >> 1);
    while (y1 != y2) {
      if (fraction >= 0) {
        x1 += stepx;
        fraction -= dy;
      }
      y1 += stepy;
      fraction += dx;
      set_pixel(x1, y1, color, dst);
    }
  }
}

/******************************************************************************/

/*
** Name: polygon_clip
** Desc: clip a polygon to pre-calculated plane distances.
*/
#ifndef __PATCH_POLYGON_CLIP__
int CODE_IN_IWRAM polygon_clip(yeti_t* yeti, vertex_t** dst, vertex_t** src, int n)
{
  vertex_t** tmp = dst;
  int isLine = (n == 2);

  if (n > 1)
  {
    vertex_t* pre;

    CACHE_ENABLE;
    for (pre = src[n - 1]; n--; pre = *src++)
    {
      if (pre->d >= 0) *dst++ = pre;

      if ((pre->d ^ src[0]->d) < 0 && (isLine ? pre->d < 0 : 1))
      {
        vertex_t* dv = *dst++ = YETI_VERTEX(yeti);
        vertex_t* sv = src[0];
        int d = pre->d - sv->d;
        int r = f2i(d < 0 ? reciprocal[-d] * -pre->d : reciprocal[d] * pre->d);

        dv->x = fixlerp(pre->x, sv->x, r);
        dv->y = fixlerp(pre->y, sv->y, r);
        dv->z = fixlerp(pre->z, sv->z, r);

        vertex_project(dv);
      }
    }
    CACHE_DISABLE;
  }
  return dst - tmp;
}
#endif

/******************************************************************************/

/*
** Name: draw_clipped_poly
** Desc: Renders a clipped poly to a screen. The texture is DDA mapped to the given polygon.
*/
#ifndef __PATCH_DRAW_CLIPPED_POLY__
void CODE_IN_IWRAM draw_clipped_poly(yeti_t* yeti, polyclip_t src, int n, int tid)
{
  int i;
  polygon_t p;

  CACHE_ENABLE;
  for (i = n; i--;)
  {
    p[i].x = src[i]->sx;
    p[i].y = src[i]->sy;
  }
  CACHE_DISABLE;

  draw_poly(p, n, tid, yeti->baffer);
}
#endif

/******************************************************************************/

/*
** Name: draw_texture
** Desc: Renders a texture to a screen. The texture is DDA mapped to the given polygon.
*/
#ifndef __PATCH_DRAW_TEXTURE__
void CODE_IN_IWRAM draw_texture(yeti_t* yeti, polyclip_t p, int n, int tid)
{
  int i;
  int norg = n;

  for (i = n; i--;)
  {
    if (
       (p[i]->x + p[i]->z) < 0 ||
       (p[i]->z - p[i]->x) < 0 ||
       (p[i]->y + p[i]->z) < 0 ||
       (p[i]->z - p[i]->y) < 0 )
    {
      polyclip_t a, b;

      for (i = n; i--; p[i]->d = p[i]->x + p[i]->z); n = polygon_clip(yeti, a, p, n); // Left
      for (i = n; i--; a[i]->d = a[i]->z - a[i]->x); n = polygon_clip(yeti, b, a, n); // Right
      for (i = n; i--; b[i]->d = b[i]->y + b[i]->z); n = polygon_clip(yeti, a, b, n); // Top
      for (i = n; i--; a[i]->d = a[i]->z - a[i]->y); n = polygon_clip(yeti, b, a, n); // Bottom

      p = b;
      
      break;
    }
  }
  if (n > 2) draw_clipped_poly(yeti, p, n, tid);
  
  else if (n == 2 && norg == 2)
    draw_line(p[0]->sx, p[0]->sy, p[1]->sx, p[1]->sy, tid, yeti->baffer);
}
#endif
/******************************************************************************/

/*
** Name: rotate_vector
** Desc: Rotates a cell vector and apply world lighting. Rotated vectors are
**       cached in map cells so they can be reused. The rotated vertex is
**       indexed by its un-rotated y position.
*/
vertex_t* CODE_IN_IWRAM rotate_vector(yeti_t* yeti, vec3_t* v)
{
  int x, y, z;
  vertex_t* rv;
  //cell_t* cell = &yeti->cells[f2i(v->z)][f2i(v->x)];

  //if (!CELL_IS_MARKED(cell))
    //CELL_MARK(cell);
  
  x = v->x - yeti->camera->x;
  y = v->y - yeti->camera->z;
  z = v->z - yeti->camera->y;

  rv    = YETI_VERTEX(yeti);
  rv->x = f2i(yeti->m[0][0] * x + yeti->m[0][1] * y + yeti->m[0][2] * z);
  rv->y = yeti->is2d ? y : f2i(yeti->m[1][0] * x + yeti->m[1][1] * y + yeti->m[1][2] * z);
  rv->z = f2i(yeti->m[2][0] * x + yeti->m[2][1] * y + yeti->m[2][2] * z);

  vertex_project(rv);

  return rv;
}

/*
** Name: draw_quad
** Desc: Renders a single quad/rectangle of a cell using the given texture
**       mapping mode and texture id. The mapping mode changes depending
**       of if this is a north-south wall, east-west wall or ceiling/floor.
*/
void CODE_IN_IWRAM draw_quad(yeti_t* yeti, int tid, quad_t quad)
{
  polyclip_t p;

  p[0] = rotate_vector(yeti, &quad[0]);
  p[1] = rotate_vector(yeti, &quad[1]);
  p[2] = rotate_vector(yeti, &quad[2]);
  p[3] = rotate_vector(yeti, &quad[3]);

  draw_texture(yeti, p, 4, tid);
}

void CODE_IN_IWRAM draw_wire(yeti_t* yeti, int tid, vec3_t A, vec3_t B)
{
  polyclip_t p;

  p[0] = rotate_vector(yeti, &A);
  p[1] = rotate_vector(yeti, &B);

  draw_texture(yeti, p, 2, tid);
}

/******************************************************************************/

/*
** Name: draw_entity_as_sprite
** Desc: Draws a entity as a 2D sprite.
*/
void CODE_IN_IWRAM draw_entity_as_sprite(entity_t* e)
{
  sprite_t sprite = e->visual.data;
  int w = sprite_width (sprite) * e->visual.width;
  int h = sprite_height(sprite) * e->visual.height;

  draw_sprite_di(e->yeti, e->rx - w, e->ry + h, e->rx + w, e->ry - h,
    e->rz, sprite, e->visual.mode);
}

/*
** Name: draw_entity_as_model
** Desc: Draws a entity 3D model at the given location.
*/
void CODE_IN_IWRAM draw_entity_as_model(entity_t* entity)
{
  s8* model = entity->visual.data;
  yeti_t* yeti = entity->yeti;
  matrix_t m;
  polyclip_t p;
  vec3_t verts[100];
  int i, j, nvertices;

  if (!model)
    return;

  model++;

  matrix_rotate_object(m, f2i(entity->r), f2i(entity->p), f2i(entity->t));
  
  for (nvertices = *model++, i = 0; i < nvertices; i++)
  {
    int u = *model++;
    int v = *model++;
    int w = *model++;
    int x = f2i(m[0][0] * u + m[0][1] * v + m[0][2] * w) + entity->x - yeti->camera->x;
    int y = f2i(m[1][0] * u + m[1][1] * v + m[1][2] * w) + entity->z - yeti->camera->z;
    int z = f2i(m[2][0] * u + m[2][1] * v + m[2][2] * w) + entity->y - yeti->camera->y;
    
    verts[i].x = f2i(yeti->m[0][0] * x + yeti->m[0][1] * y + yeti->m[0][2] * z);
    verts[i].y = yeti->is2d ? y : f2i(yeti->m[1][0] * x + yeti->m[1][1] * y + yeti->m[1][2] * z);
    verts[i].z = f2i(yeti->m[2][0] * x + yeti->m[2][1] * y + yeti->m[2][2] * z);
  }
  for (i = *model++; i--;)
  {
    int npts = *model++;
    for (j = 0; j < npts; j++)
    {
      p[j] = YETI_VERTEX(yeti);
      *((vec3_t*)&p[j][0]) = verts[*model++];
      p[j]->u = i2f(*model++);
      p[j]->v = i2f(*model++);
      p[j]->l = i2f(48);
      vertex_project(p[j]);
    }
    draw_texture(yeti, p, npts, *model++);
  }
}

/******************************************************************************/

void CODE_IN_IWRAM yeti_draw(yeti_t* yeti)
{
  cell_t* cell;
  rom_cell_t* cello;
  quad_t q, q1, q2;
  vec3_t *p, *p1, *p2, A, B;
  entity_t* e1;
  entity_t* e2;
  int i, j, x, y;
  track_segment_t* ts;
  s8 incdec = yeti->main_entity->x & 0x80?-1:1;
  u8 arccolor;

  p = (vec3_t*)&q;
  p2 = 0;


  yeti->is2d = !yeti->camera->p && !yeti->camera->r;
  
  matrix_rotate_world(yeti->m,
    -f2i(yeti->camera->r),
    -f2i(yeti->camera->p),
    -f2i(yeti->camera->t));
  
  ts = yeti->main_entity->segment + 16;
  if (ts > (track_segment_t*)&tracksegments[114])
    ts -= 115;

  for (i = 15; i >= -1; i--)
  {
    p[0].x = ts->left.x;  p[0].y = ts->left.y;  p[0].z = ts->left.z;
    p[1].x = ts->right.x; p[1].y = ts->right.y; p[1].z = ts->right.z;

    if (ts > (track_segment_t*)&tracksegments[0])
      ts--;
    else
      ts = (track_segment_t*)&tracksegments[114];

    p[3].x = ts->left.x;  p[3].y = ts->left.y;  p[3].z = ts->left.z;
    p[2].x = ts->right.x; p[2].y = ts->right.y; p[2].z = ts->right.z;

    draw_quad(yeti, /*ts == yeti->main_entity->segment ? 9 :*/ ts->color, p);

    if (ts->flags == 1)
    {
      if (i > 11)
        arccolor = 1;
      else if (i > 7)
        arccolor = 2;
      else
        arccolor = 3;

      A.x = ts->left.x; A.y = ts->left.y + 192; A.z = ts->left.z;
      B.x = ts->right.x; B.y = ts->right.y + 192; B.z = ts->right.z;
      draw_wire(yeti, arccolor, p[3], A);
      draw_wire(yeti, arccolor, p[2], B);
      draw_wire(yeti, arccolor, A, B);
    }
  }

  draw_entity_as_model(yeti->main_entity);

}

void draw_HUD(yeti_t* yeti)
{
  draw_bitstring_image(yeti,  4, 4, 9, 32, 4, (const BYTE*)digits_mask[yeti->timer[5]], digits[yeti->timer[5]]);
  draw_bitstring_image(yeti, 14, 4, 9, 32, 4, (const BYTE*)digits_mask[yeti->timer[4]], digits[yeti->timer[4]]);

  draw_bitstring_image(yeti, 25, 4, 4, 12, 4, (const BYTE*)apostrophe_mask, (const BYTE*)apostrophe);

  draw_bitstring_image(yeti, 34, 4, 9, 32, 4, (const BYTE*)digits_mask[yeti->timer[3]], digits[yeti->timer[3]]);
  draw_bitstring_image(yeti, 44, 4, 9, 32, 4, (const BYTE*)digits_mask[yeti->timer[2]], digits[yeti->timer[2]]);

  draw_bitstring_image(yeti, 54, 4, 8, 12, 4, (const BYTE*)quote_mask, (const BYTE*)quote);

  draw_bitstring_image(yeti, 64, 4, 9, 32, 4, (const BYTE*)digits_mask[yeti->timer[1]], digits[yeti->timer[1]]);
  draw_bitstring_image(yeti, 74, 4, 9, 32, 4, (const BYTE*)digits_mask[yeti->timer[0]], digits[yeti->timer[0]]);

  //2nd
  draw_bitstring_image(yeti, 171, 4, 43, 58, 8, (const BYTE*)place2nd_mask, (const BYTE*)place2nd);

  //map
  draw_bitstring_image_without_mask(yeti, 277, 4, 103, 88, 12, (const BYTE*)map1);

  //dot
  draw_bitstring_image(yeti, 277+((fixdiv(yeti->main_entity->y - 1857, 7113-1857)*26368)>>16), 2+((fixdiv(yeti->main_entity->x - 25602, 27874-25602)*11264)>>16), 4, 8, 4, (const BYTE*)mapdot_mask, (const BYTE*)mapdot);

  //speed
  draw_bitstring_image(yeti, 4, 188, 115, 64, 8, (const BYTE*)speed_mask, (const BYTE*)speed);  

  //power
  draw_bitstring_image(yeti, 147, 204, 91, 32, 4, (const BYTE*)power_mask, (const BYTE*)power);
  draw_bitstring_image_without_mask(yeti, 150, 211, yeti->main_entity->life, 12, 4, (const BYTE*)powermeter);
}

void draw_bitstring_image(yeti_t* yeti, int x, int y, int width, int bheight, int Bheight, const BYTE* mask, const BYTE* image)
{
  int offset = (x << 4) + (y >> 4);
  u32 dstOffset = (y & 0xf) << 1;

  CACHE_ENABLE;
  for (; width--;)
  {
	asm("
		st.w	r29,-4[r3]
		mov	%0,r26
		mov	0,r27
		mov	%1,r28
		mov	%2,r29
		mov	%3,r30
		.hword 0x7C09
		mov	%0,r26
		mov	0,r27
		mov	%1,r28
		mov	%2,r29
		mov	%4,r30
		.hword 0x7C08
		ld.w	-4[r3],r29
	"
	:  /* Output */
	: "r" (dstOffset), "r" (bheight), "r" (&yeti->baffer[offset]), "r" (mask), "r" (image) /* Input */
	: "r26", "r27", "r28", "r29", "r30" /* Clobber */
	);

    offset += 16;
    mask += Bheight;
    image += Bheight;
  }
  CACHE_DISABLE;
}

void draw_bitstring_image_without_mask(yeti_t* yeti, int x, int y, int width, int bheight, int Bheight, const BYTE* image)
{
  int offset = (x << 4) + (y >> 4);
  u32 dstOffset = (y & 0xf) << 1;

  CACHE_ENABLE;
  for (; width--;)
  {
	asm("
		st.w	r29,-4[r3]
		mov	%0,r26
		mov	0,r27
		mov	%1,r28
		mov	%2,r29
		mov	%3,r30
		.hword 0x7C08
		ld.w	-4[r3],r29
	"
	:  /* Output */
	: "r" (dstOffset), "r" (bheight), "r" (&yeti->baffer[offset]), "r" (image) /* Input */
	: "r26", "r27", "r28", "r29", "r30" /* Clobber */
	);

    offset += 16;
    image += Bheight;
  }
  CACHE_DISABLE;
}
