#include "game.h"

/******************************************************************************/

#define YETI_MAX_MODELS 6

YETI_ROM s8 falcon[] =
{
  1,              // Number of vertex lists (Animation Frames)
  20,              // Number of verices per list.
   33, -27, -24, // 0 Right wing bottom
  -12,   0, -12, // 1 Left main body
    0,   0,  54, // 2 Front main body
    0,  30,   0, // 3 Top main body

   12, -21, -24, // 4 Right part back left
   12,  21, -24, // 5 Right part back top
   24,  -3, -24, // 6 Right part back right
   12, -21,  12, // 7 Right part front left
   12,  21,  12, // 8 Right part front top
   24,  -3,  12, // 9 Right part front right
   18,  60, -24, // 10 Right wing top
   12,   0, -12, // 11 Right main body

  -12, -21, -24, // 12 Left part back right
  -12,  21, -24, // 13 Left part back top
  -24,  -3, -24, // 14 Left part back left
  -12, -21,  12, // 15 Left part front right
  -12,  21,  12, // 16 Left part front top
  -24,  -3,  12, // 17 Left part front left
  -18,  60, -24, // 18 Left wing top
  -33, -27, -24, // 19 Left wing bottom

  21,              // Number of faces in the polygon mesh.
                  // Each face is (NPoints, VID, U, V) + texture id.
  3, 0, 0, 0, 9, 0, 0, 6, 0, 0, 7,  // Right lower wing inner
  4, 7, 0, 0, 8, 0, 0, 5, 0, 0, 4, 0, 0, 7, // Right part inner
  3, 19, 0, 0, 14, 0, 0, 17, 0, 0, 7,  // Left lower wing inner
  4, 12, 0, 0, 13, 0, 0, 16, 0, 0, 15, 0, 0, 7, // Left part inner
  3, 1, 0, 0, 2, 0, 0, 3, 0, 0, 7, // Left body
  3, 15, 0, 0, 16, 0, 0, 17, 0, 0, 0,  // Left part front
  3, 11, 0, 0, 3, 0, 0, 2, 0, 0, 9, // Right body
  3, 7, 0, 0, 9, 0, 0, 8, 0, 0, 0, // Right part front
  3, 1, 0, 0, 3, 0, 0, 11, 0, 0, 0, // Back body
  4, 17, 0, 0, 16, 0, 0, 13, 0, 0, 14, 0, 0, 9, // Left part upper
  4, 15, 0, 0, 17, 0, 0, 14, 0, 0, 12, 0, 0, 7, // Left part lower
  4, 6, 0, 0, 5, 0, 0, 8, 0, 0, 9, 0, 0, 9, // Right part upper
  4, 4, 0, 0, 6, 0, 0, 9, 0, 0, 7, 0, 0, 7, // Right part lower
  3, 8, 0, 0, 10, 0, 0, 5, 0, 0, 9, // Right upper wing inner
  3, 16, 0, 0, 18, 0, 0, 13, 0, 0, 7, // Left upper wing outer
  3, 13, 0, 0, 18, 0, 0, 16, 0, 0, 9, // Left upper wing inner
  3, 5, 0, 0, 10, 0, 0, 8, 0, 0, 7, // Right upper wing outer
  3, 0, 0, 0, 6, 0, 0, 9, 0, 0, 8, // Right lower wing outer
  3, 19, 0, 0, 17, 0, 0, 14, 0, 0, 8, // Left lower wing outer
  3, 4, 0, 0, 5, 0, 0, 6, 0, 0, 0, // Right part back
  3, 12, 0, 0, 14, 0, 0, 13, 0, 0, 0 // Left part back
};

/******************************************************************************/

int friction(int a, int b)
{
  return a - f2i((b = a * b) < 0 ? b - FIXCEIL : b + FIXCEIL);
}

void entity_friction(entity_t* e, int amount)
{
  e->xx = friction(e->xx, amount);
  e->yy = friction(e->yy, amount);
}

void entity_move_forward(entity_t* e)
{
  e->xx = (fixsin(f2i(e->t)) * e->v) >> 6;
  e->yy = (fixcos(f2i(e->t)) * e->v) >> 6;
}

void entity_move_backwards(entity_t* e)
{
  e->xx = (fixsin(f2i(e->t)) * e->v) >> 6;
  e->yy = (fixcos(f2i(e->t)) * e->v) >> 6;
}

void entity_strafe_right(entity_t* e)
{
  e->xx += fixsin(f2i(e->t)+500) >> 4;
  e->yy += fixcos(f2i(e->t)+500) >> 4;
}

void entity_strafe_left(entity_t* e)
{
  e->xx -= fixsin(f2i(e->t)+500) >> 4;
  e->yy -= fixcos(f2i(e->t)+500) >> 4;
}

void entity_turn_right(entity_t* e)
{
  e->tt += i2f(25);
}

void entity_turn_left(entity_t* e)
{
  e->tt -= i2f(25);
}

void entity_turn_towards(entity_t* e, int x, int y)
{
  int x1 = x - (e->x + fixsin(f2i(e->t + i2f(100))));
  int y1 = y - (e->y + fixcos(f2i(e->t + i2f(100))));
  int x2 = x - (e->x + fixsin(f2i(e->t - i2f(100))));
  int y2 = y - (e->y + fixcos(f2i(e->t - i2f(100))));

  if ((x1 * x1 + y1 * y1) < (x2 * x2 + y2 * y2))
  {
    entity_turn_right(e);
  }
  else
  {
    entity_turn_left(e);
  }
}

void entity_kill(entity_t* e)
{
  yeti_t* yeti = e->yeti;

  e->ontick = 0;
  e->onhit  = 0;
  e->visual.data = 0;
  e->radius = 0;
}

/******************************************************************************/

void entity_default(entity_t* e, int isjumping, int iscrawling)
{
  yeti_t* yeti = e->yeti;

  //e->xx = friction(e->xx, 30);
  //e->yy = friction(e->yy, 30);
  //e->tt = friction(e->tt, 80);

  //e->r = friction(e->r, 20);
  //e->p = friction(e->p, 20);

  {
    rom_cell_t* cell = &yeti->curr_map->cells[f2i(e->y)][f2i(e->x)];
    int bot = cell->bot + (iscrawling ? yeti->game.crawling : yeti->game.walking);
    int top = cell->top - yeti->game.ceiling;

    if (e->z <= bot)
    {
      e->z  -= ((e->z - bot) >> 2);
      e->zz -= (e->zz >> 2);
      if (isjumping && e->z < top)
      {
        e->z = bot;
        e->zz += yeti->game.jumping;
      }
    }
    else
    {
      e->zz += yeti->game.gravity;
    }

    if (e->z >= top)
    {
      e->z -= ((e->z - top) >> 2);
    }
  }
  if (e->life < 0) entity_kill(e);
}

void bullet_collision(entity_t* e1, entity_t* e2)
{
  yeti_t* yeti = e1->yeti;

  if (e2 != yeti->camera)
  {
    e2->life -= 10;
    entity_kill(e1);
  }
}

void bullet_behaviour(entity_t* e)
{
  e->onhit = bullet_collision;
  if (!e->xx || !e->yy)
  {
    entity_kill(e);
  }
}

void shoot_bullet(entity_t* e)
{
  yeti_t* yeti = e->yeti;

  entity_t* b = YETI_BULLET(yeti);

  b->t  = e->t;
  b->p  = e->p;
  b->xx = (fixsin(f2i( e->t))) + (e->xx);
  b->yy = (fixcos(f2i( e->t))) + (e->yy);
  b->zz = (fixsin(f2i(-e->p))) + (e->zz);
  b->x  = e->x;
  b->y  = e->y;
  b->z  = e->z;
  b->visual.data = (void*)spr_ball1;
  b->visual.width = 1;
  b->visual.height = 1;
  b->visual.mode = 1;
  b->radius = 16;
  b->ontick = bullet_behaviour;
}

void camera_behaviour(entity_t* e)
{
  yeti_t* yeti = e->yeti;
  h_cell_t *hc0, *hc1, *hc2, *hc3;
  int x1,x2,xin,yin;
  track_segment_t* nxtseg;

  if (yeti->keyboard.pad & K_LT) e->p -= i2f(20);
  if (yeti->keyboard.pad & K_RT) e->p += i2f(20);

  if (yeti->camera_zoom_to > yeti->camera_zoom) yeti->camera_zoom += 20;
  else if (yeti->camera_zoom_to < yeti->camera_zoom) yeti->camera_zoom -= 20;

  //e->x = yeti->main_entity->x;
  //e->y = yeti->main_entity->y - yeti->camera_zoom;//180; ////320
  e->t = yeti->main_entity->t;
  e->x = yeti->main_entity->x - fixmul(yeti->camera_zoom, fixsin(f2i(e->t)));
  e->y = yeti->main_entity->y - fixmul(yeti->camera_zoom, fixcos(f2i(e->t)));

  nxtseg = e->segment + 1;
  if (nxtseg > (track_segment_t*)&tracksegments[114])
    nxtseg = (track_segment_t*)&tracksegments[0];

  if (ABS(e->segment->right.x - e->segment->left.x) < ABS(e->segment->right.z - e->segment->left.z)) {
    xin = fixdiv(e->y + e->yy - e->segment->left.z, e->segment->right.z - e->segment->left.z);
    yin = fixdiv(e->x + e->xx - e->segment->left.x, nxtseg->left.x - e->segment->left.x);
  }
  else {
    xin = fixdiv(e->x + e->xx - e->segment->left.x, e->segment->right.x - e->segment->left.x);
    yin = fixdiv(e->y + e->yy - e->segment->left.z, nxtseg->left.z - e->segment->left.z);
  }
  
  x1 = e->segment->left.y + (yin * (nxtseg->left.y - e->segment->left.y) >> 8);
  x2 = e->segment->right.y + (yin * (nxtseg->right.y - e->segment->right.y) >> 8);
  e->z = 110 + x1 + (xin * (x2 - x1) >> 8);

  /*if (yeti->keyboard.pad & K_STA) {
    e->y = yeti->main_entity->y;
    e->z = 1500;
    e->p = i2f(512);
  }
  else if (yeti->keyboard.pad & K_SEL) {
    e->y = yeti->main_entity->y;
    e->x -= 256;
    e->z = 110;
    e->t = i2f(512);
  }
  else*/
    e->p = i2f(40);
}

void box_behaviour(entity_t* e)
{
  e->t += i2f(25);
}

void player_collision(entity_t* e1, entity_t* e2);

void player_rent_hit(entity_t* e, u16 rom_entity_ID, s32 x_base)
{
  yeti_t* yeti = e->yeti;
}

void timed_explosion_behaviour(entity_t* e)
{
  e->t += i2f(200);
  e->life -= 15;

  if (e->life < 0)
    entity_kill(e);
}

void vehicle_behaviour(entity_t* e)
{
  yeti_t* yeti = e->yeti;
  h_cell_t *hc0, *hc1, *hc2, *hc3;
  int x1,x2,y1,y2,xin,yin;
  u8 rot = 0;
  track_segment_t* nxtseg;
  
  if (yeti->keyboard.pad & K_LL)
  {
    e->t -= i2f(30);
    entity_move_forward(e);
    if (e->rc > -30000)
      e->rc -= 10000;

  }
  else if (yeti->keyboard.pad & K_LR)
  {
    e->t +=  i2f(30);
    entity_move_forward(e);
    if (e->rc < 30000)
      e->rc += 10000;
  }
  else if (e->rc > 0)
    e->rc -= 10000;
  else if (e->rc < 0)
    e->rc += 10000;
  if (yeti->keyboard.pad & K_A && e->v < 8) {e->v += 1; entity_move_forward(e);} else
  if (yeti->keyboard.pad & K_B && e->v > 0) {e->v -= 1; entity_move_forward(e);}

  /*if ((yeti->keyboard.pad & K_SEL) && !(yeti->keyboard._pad & K_SEL) ) {
    yeti->currModel++;
    if (yeti->currModel >= YETI_MAX_MODELS)
      yeti->currModel = 0;
    e->visual.data = (s8*)models[yeti->currModel];
  }*/

  /*if ((yeti->keyboard.pad & K_LT) && !(yeti->keyboard._pad & K_LT) )
    yeti->mapon = yeti->mapon ? 0 : 1;

  if ((yeti->keyboard.pad & K_RT) && !(yeti->keyboard._pad & K_RT) ) {
    entity_t* et = yeti_entity(yeti, e->x + e->radius + 10, e->y, 1000, walking_behaviour);
    et->visual.data = &player;
    et->t = e->t;
    et->radius = 9;
    et->onhit = player_collision;
    et->onBGhit = player_rent_hit;
    yeti->main_entity = et;
    yeti->camera_zoom_to = 180;
    e->ontick = vehicle_behaviour_auto;
  }*/

  nxtseg = e->segment + 1;
  if (nxtseg > (track_segment_t*)&tracksegments[114])
    nxtseg = (track_segment_t*)&tracksegments[0];

  if (ABS(e->segment->right.x - e->segment->left.x) < ABS(e->segment->right.z - e->segment->left.z)) {
    xin = fixdiv(e->y + e->yy - e->segment->left.z, e->segment->right.z - e->segment->left.z);
    yin = fixdiv(e->x + e->xx - e->segment->left.x, nxtseg->left.x - e->segment->left.x);
    rot = 1;
  }
  else {
    xin = fixdiv(e->x + e->xx - e->segment->left.x, e->segment->right.x - e->segment->left.x);
    yin = fixdiv(e->y + e->yy - e->segment->left.z, nxtseg->left.z - e->segment->left.z);
  }
  
  x1 = e->segment->left.y + (yin * (nxtseg->left.y - e->segment->left.y) >> 8);
  x2 = e->segment->right.y + (yin * (nxtseg->right.y - e->segment->right.y) >> 8);

  y1 = e->segment->left.y + (xin * (e->segment->right.y - e->segment->left.y) >> 8);
  y2 = nxtseg->left.y + (xin * (nxtseg->right.y - nxtseg->left.y) >> 8);

  e->z = x1 + (xin * (x2 - x1) >> 8) - ((s8*)e->visual.data)[3];
  e->p = ((rot ? y1 - y2 : x1 - x2) * fixsin(f2i(e->t)) + (rot ? x1 - x2 : y1 - y2) * fixcos(f2i(e->t))) << 1;
  e->r = ((rot ? x1 - x2 : y2 - y1) * fixsin(f2i(e->t)) + (rot ? y1 - y2 : x2 - x1) * fixcos(f2i(e->t))) << 1;

  e->r += e->rc;
}

u8 test = 0;
void player_collision(entity_t* e1, entity_t* e2)
{
  yeti_t* yeti = e1->yeti;

  test++;

  yeti->camera_zoom_to = 320;
}

void guard_collision(entity_t* e1, entity_t* e2)
{
//
}

void guard_behaviour(entity_t* e)
{
  /*yeti_t* yeti = e->yeti;

  e->visual.data   = sprites[(yeti->tick >> 2) & 3];
  e->visual.width  = 8;
  e->visual.height = 12;
  e->radius = 200;
  e->onhit = guard_collision;

  entity_turn_towards(e, yeti->camera->x, yeti->camera->y);

  entity_move_forward(e);
  e->xx = friction(e->xx, 50);
  e->yy = friction(e->yy, 50);

  entity_default(e, FALSE, FALSE);*/
}

void game_init(yeti_t* yeti)
{
  int x, y;
  entity_t* enti;

  yeti->game.gravity = -16;      // Gravity force.
  yeti->game.jumping = 80;       // Jumping force.
  yeti->game.crawling = i2f(1);  // Crawling height.
  yeti->game.walking = 130;      // Walking height.
  yeti->game.ceiling = 200;      // Crouch distance.

  yeti_load_map(yeti, 0, 0);

  //yeti->camera->ontick = camera_behaviour;
  yeti->camera->segment = (track_segment_t*)&tracksegments[0];
  yeti->camera->p = i2f(40);
  yeti->camera_zoom = 600;
  yeti->camera_zoom_to = 180;

  enti = yeti_entity(yeti, i2f(100)+127, i2f(17)+61, i2f(0), vehicle_behaviour);
  enti->visual.data = &falcon;
  //enti->t = i2f(1280);
  enti->radius = 33;
  enti->life = 85;
  enti->segment = (track_segment_t*)&tracksegments[0];
  yeti->main_entity = enti;
  yeti->currModel = 1;

  enti = yeti_entity(yeti, i2f(112)+100, i2f(16)+150, i2f(0), box_behaviour);
  enti->visual.data = &falcon;
  enti->t = i2f(880);
  enti->radius = 33;
}

void game_draw_item(yeti_t* yeti, int x, int y, int u, int v)
{
  /*
  u <<= 6;
  v <<= 6;
  stretch_sprite_di(yeti,
    -i2f(1) - x, y + i2f(1), i2f(1) - x, y - i2f(1), i2f(8)+50, spr_items,
    u, v, u + 64, v + 64, 1);
  */
}

void game_draw(yeti_t* yeti)
{
  yeti_draw(yeti);
  draw_HUD(yeti);
}

void game_tick(yeti_t* yeti)
{
  //char teckst[20];
  //posprintf(teckst, "%d", (s32)((fixdiv(yeti->main_entity->y - 1857, 7113-1857)*26368)>>16));
  //vbTextOut(1, 0, 6, teckst);
  yeti_tick(yeti);

  yeti->timer[0] += 3;

  if (yeti->timer[0] > 9) {
    yeti->timer[0] -= 10;
    yeti->timer[1]++;

    if (yeti->timer[1] > 9) {
      yeti->timer[1] = 0;
      yeti->timer[2]++;

      if (yeti->timer[2] > 9) {
        yeti->timer[2] = 0;
        yeti->timer[3]++;

        if (yeti->timer[3] > 5) {
          yeti->timer[3] = 0;
          yeti->timer[4]++;

          if (yeti->timer[4] > 9) {
            yeti->timer[4] = 0;
            yeti->timer[5]++;

            if (yeti->timer[5] > 9)
              yeti_memset(yeti->timer, 9, 6);
          }
        }
      }
    }
  }
}

void game_loop(yeti_t* yeti)
{
  game_tick(yeti);
  game_draw(yeti);
}

