#include "yeti.h"

#include "include\world.h"
#include "include\mem.h"
#include "include\object.h"
#include "include\joypad.h"
#include "include\asm.h"
#include "include\interrupts.h"
#include "include\timer.h"

#include "chgun.h"
#include "mpgun.h"
#include "mpstars.h"
#include "chBoomBox.h"
#include "font_pc_bios.h"

#include "bgadjustment.h"
#include "chadjustment.h"
#include "mpprecaution.h"


static HWORD objNumbers[] = {
//	X	L/RON & P	Y	Pal/HFlp/VFlp/Char
	370,	0xFFFC,		0,	1014,	// ental
	363,	0xFFFC,		0,	1014,	// tiotal
	356,	0xFFFC,		0,	1014,	// hundratal
	349,	0xFFFC,		0,	1014,	// tusental
	342,	0xFFFC,		0,	1014,	// tiotusental

	198,	0xC000,		106,	288,	// top left
	206,	0xC000,		106,	289,	// top right
	198,	0xC000,		114,	305,	// bottom left
	206,	0xC000,		114,	306,	// bottom right

	187,	0xC000,		106,	288,	// top left
	195,	0xC000,		106,	289,	// top right
	187,	0xC000,		114,	305,	// bottom left
	195,	0xC000,		114,	306,	// bottom right

	176,	0xC000,		106,	288,	// top left
	184,	0xC000,		106,	289,	// top right
	176,	0xC000,		114,	305,	// bottom left
	184,	0xC000,		114,	306	// bottom right
};

IN_EWRAM yeti_t yeti;

int min(int a, int b)
{
  return MIN(a, b);
}

int max(int a, int b)
{
  return MAX(a, b);
}

void vb_init(void)
{ u16* txtinit;
  u16* const param = (u16*)0x0003C000;

  //Precaution screen
  copymem((void*)CharSeg3,(void*)FONT,512*16);
  //copymem((void*)BGMap(0),(void*)MPPRECAUTION,8192); // 64*64*2
  //vbSetWorld(31, WRLD_ON, 0, 0, 0, 0, 0, 0, 384, 224);
  //vbSetWorld(30,WRLD_END, 0, 0, 0, 0, 0, 0, 0, 0);

  //vbDisplayOn();
  //vbFXFadeIn(1);
  //while (!(vbReadPad()&(K_ANY)));
  //vbFXFadeOut(1);

  //IPD screen
  //copymem((void*)CharSeg0, (void*)CHADJUSTMENT, 256*16);
  //copymem((void*)BGMap(0), (void*)BGADJUSTMENT, 450*16);
  //vbSetWorld(31, WRLD_LON|WRLD_OVR, 0, 0, 0, 0, 0, 0, 384, 224);
  //vbSetWorld(30, WRLD_RON|WRLD_OVR, 0, 0, 0, 0, 0, 224, 384, 224);
  //vbSetWorld(29, WRLD_END, 0, 0, 0, 0, 0, 0, 0, 0);

  //vbFXFadeIn(1);
  //while (!(vbReadPad()&(K_ANY)));
  //vbFXFadeOut(1);

  //Title screen
  //copymem((void*)CharSeg0,(void*)CHTITLE,512*16);
  //copymem((void*)BGMap(0),(void*)MPTITLE,8192); // 64*64*2
  //vbSetWorld(31, WRLD_ON, 0, 0, 0, 0, 0, 0, 384, 224);
  //vbSetWorld(30,WRLD_END, 0, 0, 0, 0, 0, 0, 0, 0);

  //vbFXFadeIn(1);
  //while (!(vbReadPad()&(K_ANY)));
  //vbFXFadeOut(1);

  //Game setup
  copymem((void*)CharSeg0,(void*)CHHUD,512*16);
  copymem((void*)BGMap(0),(void*)MPHUD,8192); // 64*64*2
  copymem((void*)BGMap(2),(void*)MPSTARS,8192);
  copymem((void*)OAMBase, (void*)objNumbers, 136);
  //copymem((void*)CharSeg1,(void*)CHBOOMBOX,512*16);

  vbSetWorld(31,WRLD_ON|2,0,6,0,0,0,0,384,224); // Stars
  vbSetWorld(30,WRLD_ON,120,3,66,392,0,464,27,19); // Moon
  vbSetWorld(29,0,0,0,0,0,0,0,10,10); // Ocean
  vbSetWorld(28,/*WRLD_ON|*/WRLD_HBIAS|WRLD_OVR,120,3,98,392,0,492,36,19); // Moon reflection
  vbSetWorld(27,/*WRLD_ON*/0,0,0,189,0,0,40,384,35); // HUD mask
  vbSetWorld(26,/*WRLD_ON*/0,32,-12,189,0,0,0,320,40); // HUD
  vbSetWorld(25,/*WRLD_ON|*/WRLD_OBJ,0,0,0,0,0,0,0,0);
  vbSetWorld(24,/*WRLD_ON|*/1,0,0,0,0,0,0,384,224); // Debug text
  vbSetWorld(23,WRLD_END, 0, 0, 0, 0, 0, 0, 0, 0);
  for (txtinit = &BGMM[0x1000]; txtinit != &BGMM[0x2000]; txtinit++)
    *txtinit = 1536;

  WA[28].param = (u32)param;
  WA[28].ovr = 3952;

  VIP_REGS[SPT3] = 4;
  VIP_REGS[SPT2] = -1;
  VIP_REGS[SPT1] = -1;
  VIP_REGS[SPT0] = -1;

  vbDisplayOn();
  vbDisplayShow();

  VIP_REGS[GPLT1] = 0x90;

  HW_REGS[WCR] = 1;
}

void vb_flip(void)
{
  u32* temp;
  static u16 pgflip = XPBSY1;

  yeti.keyboard.pad = vbReadPad();

  // Flip the frame buffers.

  pgflip = ~pgflip & 0x000C;
  VIP_REGS[XPCTRL] = VIP_REGS[XPSTTS] | XPEN; // sets the XPEN bit to refresh the screen
  while (!(VIP_REGS[XPSTTS] & pgflip)); // makes sure to freeze screen on framebuffer 1
  VIP_REGS[XPCTRL] = VIP_REGS[XPSTTS] & ~XPEN; // turn screen refreshing off again

  temp = yeti.scrin;
  yeti.scrin = yeti.baffer;
  yeti.baffer = temp;
}

int main()
{
  s32 camXOffset, camYOffset;

  vb_init();
  yeti_init(&yeti, (framebuffer_t*) 0x00000000, (framebuffer_t*) 0x00000000, (texture_t*)textures, (color_t*)palette, (lut_t*)lua);
  yeti.scrin  = (u32*)0x00000000;
  yeti.baffer = (u32*)0x00008000;
  game_init(&yeti);

  while (1/*!yeti.keyboard.select*/)
  {
    game_tick(&yeti);

    WA[29].gy = 91 + (yeti.camera->z - 110 >> 4);
    WA[29].my = WA[29].gy & 1 ? 289 : 288;
    WA[31].gy = WA[29].gy - 100;
    WA[31].mx = (f2i(yeti.camera->t) & 2047) >> 1;
    WA[30].gx = 180 + f2i(fixcos(yeti.tick) * 220);
    WA[30].gy = WA[29].gy + 132 + f2i(fixsin(yeti.tick) * 220);
    WA[28].gx = WA[30].gx;
    WA[28].gy = WA[29].gy + (WA[29].gy-WA[30].gy) - 20;

    camXOffset = fixsin(f2i(yeti.camera->t)+512) >> 4;
    camYOffset = fixcos(f2i(yeti.camera->t)+512) >> 4;

    yeti.camera->x -= camXOffset;
    yeti.camera->y -= camYOffset;

    game_draw(&yeti);

    yeti.camera->x += camXOffset << 1;
    yeti.camera->y += camYOffset << 1;

    yeti.baffer += 0x4000;
    game_draw(&yeti);
    yeti.baffer -= 0x4000;

    yeti.camera->x -= camXOffset;
    yeti.camera->y -= camYOffset;

    vb_flip();
  }
  return 0;
}
