#ifndef __VIEWPORTS_H__
#define __VIEWPORTS_H__

/******************************************************************************/

/*
** Some Porting Notes:
**
** - The standard pixel format is 5:5:5. There is a function in yeti.c which
**   will copy the viewport to a video display that supports a different pixel
**   format.
** - Define YETI_ROM as const for all ROM based ports.
** - You can patch functions by using the __PATCH?????__ macros.
** - Make sure you define the platform macro in your IDE/make scripts.
** - Viewports larger than 320x240 are not recommended unless you patch the
**   default texture mapper. The standard DDA code is limited to small displays.
**   Look at the OpenGL port for how todo this.
**
*/

/******************************************************************************/

#ifdef __GBA__

#define YETI_RGB555
#define YETI_LOW_QUALITY_WALLS

#define YETI_ROM const

#define YETI_VIEWPORT_INTERVAL 14
#define YETI_VIEWPORT_INTERVAL_ERROR 7

#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 160
#define YETI_VIEWPORT_Y2 128
#define YETI_VIEWPORT_WIDTH 160
#define YETI_VIEWPORT_HEIGHT 128

//#define __PATCH_DRAW_POLY__

#define CODE_IN_IWRAM __attribute__ ((section (".iwram"), long_call))
#define CODE_IN_ROM __attribute__ ((section (".text"), long_call))
#define IN_IWRAM __attribute__ ((section (".iwram")))
#define IN_EWRAM __attribute__ ((section (".ewram")))

#endif

/******************************************************************************/

#ifdef __SERIES60__

#define YETI_RGB444
#define YETI_LOW_QUALITY_WALLS

#define YETI_ROM const

#define YETI_VIEWPORT_INTERVAL 35
#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __GP32__

#define YETI_RGB565
#define YETI_LOW_QUALITY_WALLS

#define YETI_ROM

#define YETI_VIEWPORT_INTERVAL 35
#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __POCKETPC__

#define YETI_RGB565
#define YETI_LOW_QUALITY_WALLS
#define YETI_ROM

#define YETI_VIEWPORT_INTERVAL 35
#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __SVGALIB__

#define YETI_RGB555
#define YETI_ROM

#define YETI_VIEWPORT_INTERVAL 35
#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __GAMECUBE__

#define YETI_RGB565
#define YETI_ROM

#define YETI_VIEWPORT_INTERVAL 35
#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __WIN32__

#define YETI_RGB555
#define YETI_ROM 

#define YETI_VIEWPORT_INTERVAL 35
#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240
#define YETI_VIDEO_LUT

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __MSDOS__

#define YETI_RGB555
#define YETI_ROM

#define YETI_VIEWPORT_INTERVAL 35
#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __PSONE__

#define YETI_RGB555
#define YETI_ROM

#define YETI_VIEWPORT_INTERVAL 3
#define YETI_VIEWPORT_INTERVAL_ERROR 1

#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 320
#define YETI_VIEWPORT_Y2 240
#define YETI_VIEWPORT_WIDTH 320
#define YETI_VIEWPORT_HEIGHT 240

#define __PATCH_DRAW_CLIPPED_POLY__

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#ifdef __VIRTUALBOY__

#define YETI_RGB555
#define YETI_LOW_QUALITY_WALLS

#define YETI_ROM const

#define YETI_VIEWPORT_INTERVAL 14
#define YETI_VIEWPORT_INTERVAL_ERROR 7

#define YETI_VIEWPORT_X1 0
#define YETI_VIEWPORT_Y1 0
#define YETI_VIEWPORT_X2 384
#define YETI_VIEWPORT_Y2 200
#define YETI_VIEWPORT_WIDTH 384
#define YETI_VIEWPORT_HEIGHT 200

#define CODE_IN_IWRAM
#define CODE_IN_ROM
#define IN_IWRAM
#define IN_EWRAM

#endif

/******************************************************************************/

#endif
