#include "yeti.h"
#include "posprintf.h"

extern const s8 player[];
extern const s8 swimmer[];
extern const s8 rocket_boat[];

void vbTextOut(u16 bgmap, u16 col, u16 row, char *t_string)
/* The font must reside in Character segment 3 */
{
  u16* const BGMM = (u16*)0x00020000;	
  u16 i = 0;
  u16 pos = row * 64 + col;

  while(t_string[i])
  {
    BGMM[(0x1000 * bgmap) + pos + i] = (u16)t_string[i] + 0x600;
    i++;
  }
}

//char teckst[20];
//posprintf(teckst, "%3d", (s16)(hc->entityID));
//vbTextOut(1, (j+256>>8)*5, (i+256>>8)+1, teckst);

/******************************************************************************/

/*
** Name: cell_init
** Desc: Initialize a yeti cell to its defaults.
*/
void cell_init(cell_t* cell, int issolid)
{
  /*CLEARMEM(cell);
  
  if (issolid)
  {
    cell->top = i2f(0);
    cell->bot = i2f(0);
  }
  else
  {
    cell->top = i2f(4);
    cell->bot = i2f(0);
  }
#ifdef __YETI_EDITOR__
  cell->tos = cell->top;
  cell->bos = cell->bot;
#endif
  cell->lit = YETI_LIGHT_MAX;
  cell->ttx = 7;
  cell->wtx = 0;
  cell->btx = 28;*/
}

/*
** Name: yeti_entity
** Desc: Creates a new entity and returns a entity pointer.
*/
entity_t* yeti_entity(yeti_t* yeti, int x, int y, int z, entity_behaviour_t behaviour)
{
  int i;
  entity_t* e;
  for (i = 0; i < YETI_ENTITY_MAX; i++) {
    e = &yeti->entities[i];
    if (e->parent == e)
      break;
  }

  if (e->parent != e)
    return 0;

  CLEARMEM(e);
  
  e->x = x;
  e->y = y;
  e->z = z;
  e->ontick = behaviour;
  //e->radius = i2fdiv2(1);
  e->life = 127;
  e->yeti = yeti;

  return e;
}

int entity_collision(yeti_t* yeti, entity_t* e, vec2_t v, vec2_t* b)
{
  s16 i, j;
  u16* cellidx;
  vec2_t vb;
  h_cell_t* hc;
  s32 xPmin, xPmax, yPmin, yPmax, xQmin, xQmax, yQmin, yQmax;
  int c = fixcos(f2i(e->t)), s = fixsin(f2i(e->t));
  s32 xPdif, yPdif, cxPdf, sxPdf, cyPdf, syPdf;

  xPdif = ((s8*)e->visual.data)[2];
  yPdif = ((s8*)e->visual.data)[4] >> 1;
  cxPdf = c * xPdif;
  sxPdf = s * xPdif;
  cyPdf = c * yPdif;
  syPdf = s * yPdif;

  //Compute the bounding box coordinates for P.
  if (c > 0)
    if (s > 0)
    {
      xPmin = -( xPmax = f2i(cxPdf + syPdf) );
      yPmin = -( yPmax = f2i(cyPdf + sxPdf) );
    }
    else //s <= 0
    {
      xPmin = -( xPmax = f2i(cxPdf - syPdf) );
      yPmin = -( yPmax = f2i(cyPdf - sxPdf) );
    }
  else //c <= 0
    if (s > 0)
    {
      xPmax = -( xPmin = f2i(cxPdf - syPdf) );
      yPmax = -( yPmin = f2i(cyPdf - sxPdf) );
    }
    else //s <= 0
    {
      xPmax = -( xPmin = f2i(cxPdf + syPdf) );
      yPmax = -( yPmin = f2i(cyPdf + sxPdf) );
    }

  /*for (i = 0; i < 2; i++)
  {
    vb.x = v.x + b[i].x;
    vb.y = v.y + b[i].y;
    hc = (h_cell_t*)&huntercells[yeti->grid->cells[f2i(vb.y)][f2i(vb.x)]];
    //White border
    if (hc->color == 9)
      return 1;
    //Land
    if (e->visual.data == &rocket_boat)
    {
      switch (hc->style)
      {
        case 4: if (hc->color != 1) return 1; break;
        case 0: if ((vb.x & 0xff) > (vb.y & 0xff)) return 1; break;
        case 1: if ((vb.x & 0xff) < 255-(vb.y & 0xff)) return 1; break;
        case 2: if ((vb.x & 0xff) < (vb.y & 0xff)) return 1; break;
        case 3: if ((vb.x & 0xff) > 255-(vb.y & 0xff)) return 1; break;
      }
    }
    //Water
    else if (e->visual.data != &player && e->visual.data != &swimmer)
    {
      if (hc->color == 1)
        return 1;
      switch (hc->style)
      {
        case 4: break;
        case 0: if ((vb.x & 0xff) < (vb.y & 0xff)) return 1; break;
        case 1: if ((vb.x & 0xff) > 255-(vb.y & 0xff)) return 1; break;
        case 2: if ((vb.x & 0xff) > (vb.y & 0xff)) return 1; break;
        case 3: if ((vb.x & 0xff) < 255-(vb.y & 0xff)) return 1; break;
      }
    }
  }*/

  /*cellidx = &yeti->grid->cells[f2i(v.y)+1][f2i(v.x)-1];

  for (i = -256; i < 512; i+=256, cellidx-=131)
  {
    for (j = -256; j < 512; j+=256, cellidx++)
    {
      hc = (h_cell_t*)&huntercells[*cellidx];

    if (hc->entityID >= 0)
    {
      rom_entity_t* prent = (rom_entity_t*)&hunterentities[hc->entityID];

      do
      {
        if (yeti->building_strength[prent - hunterentities] == 0)
          continue;

        //xQmin = (vb.x & 0xffffff00) + prent->x - ((s8*)prent->model)[2] - v.x - f2i(syPdf);
        xQmin = prent->x - ((s8*)prent->model)[2] - (v.x & 0xff) + j - f2i(syPdf);
        xQmax = xQmin + (((s8*)prent->model)[2] << 1);
        //yQmin = (vb.y & 0xffffff00) + prent->y - ((s8*)prent->model)[4] - v.y - f2i(cyPdf);
        yQmin = prent->y - ((s8*)prent->model)[4] - (v.y & 0xff) - i - f2i(cyPdf);
        yQmax = yQmin + (((s8*)prent->model)[4] << 1);

        if ( xPmin > xQmax || xQmin > xPmax || yPmin > yQmax || yQmin > yPmax )
          continue;

        if (c > 0)
        {
          if (s > 0)
          {
            if (!(f2i(c*xQmax + s*yQmax) < -xPdif ||
                  f2i(c*xQmin + s*yQmin) >  xPdif ||
                  f2i(c*yQmax - s*xQmin) < -yPdif ||
                  f2i(c*yQmin - s*xQmax) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities, (v.x & 0xffffff00) + j);
              return 1;
            }
          }
          else //s <= 0
            if (!(f2i(c*xQmax + s*yQmin) < -xPdif ||
                  f2i(c*xQmin + s*yQmax) >  xPdif ||
                  f2i(c*yQmax - s*xQmax) < -yPdif ||
                  f2i(c*yQmin - s*xQmin) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities, (v.x & 0xffffff00) + j);
              return 1;
            }
        }
        else //c <= 0
        {
          if (s > 0)
          {
            if (!(f2i(c*xQmin + s*yQmax) < -xPdif ||
                  f2i(c*xQmax + s*yQmin) >  xPdif ||
                  f2i(c*yQmin - s*xQmin) < -yPdif ||
                  f2i(c*yQmax - s*xQmax) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities, (v.x & 0xffffff00) + j);
              return 1;
            }
          }
          else //s <= 0
            if (!(f2i(c*xQmin + s*yQmin) < -xPdif ||
                  f2i(c*xQmax + s*yQmax) >  xPdif ||
                  f2i(c*yQmin - s*xQmax) < -yPdif ||
                  f2i(c*yQmax - s*xQmin) >  yPdif)) {
              if (e->onBGhit)
                e->onBGhit(e, prent - hunterentities, (v.x & 0xffffff00) + j);
              return 1;
            }
        }

      } while ((prent++)->flags > 127);
    }
    }
  }*/
  e->x = v.x;
  e->y = v.y;
  return 0;
}

void entity_tick(entity_t* e)
{
  if (e->ontick)
    e->ontick(e);
}

/*
** Name: entity_motion
** Desc: Simple map to entity motion & collision.
*/
void entity_motion(entity_t* e)
{
  yeti_t* yeti = e->yeti;

  int dy, dx, stepx, stepy, c1, c2, c3, c4;
  vec2_t old, v, target, front, side;
  vec2_t bumper[2];
  track_segment_t* nxtseg;

  old.x = v.x = e->x;
  old.y = v.y = e->y;
  dy = e->yy;
  dx = e->xx;
  target.x = v.x + dx;
  target.y = v.y + dy;
  front.x = f2i(fixsin(f2i(e->t)) * ((s8*)e->visual.data)[4]);
  front.y = f2i(fixcos(f2i(e->t)) * ((s8*)e->visual.data)[4]);
  side.x = f2i(fixsin(f2i(e->t)+512) * ((s8*)e->visual.data)[2]);
  side.y = f2i(fixcos(f2i(e->t)+512) * ((s8*)e->visual.data)[2]);
  bumper[0].x = front.x + side.x;
  bumper[0].y = front.y + side.y;
  bumper[1].x = front.x - side.x;
  bumper[1].y = front.y - side.y;

  if (dy < 0) { dy = -dy;  stepy = -1; } else { stepy = 1; }
  if (dx < 0) { dx = -dx;  stepx = -1; } else { stepx = 1; }
  dy <<= 1;
  dx <<= 1;

  if (dx > dy) {
    int fraction = dy - (dx >> 1);
    while (v.x != target.x) {
      if (fraction >= 0) {
        v.y += stepy;
        fraction -= dx;
      }
      v.x += stepx;
      fraction += dy;
      if (entity_collision(yeti, e, v, bumper))
        break;
    }
  } else {
    int fraction = dx - (dy >> 1);
    while (v.y != target.y) {
      if (fraction >= 0) {
        v.x += stepx;
        fraction -= dy;
      }
      v.y += stepy;
      fraction += dx;
      if (entity_collision(yeti, e, v, bumper))
        break;
    }
  }

  nxtseg = e->segment + 1;
  if (nxtseg > (track_segment_t*)&tracksegments[114])
    nxtseg = (track_segment_t*)&tracksegments[0];

  //Collision with left track edge
  c1 = fixmul(nxtseg->left.x - e->segment->left.x, old.y - nxtseg->left.z) - fixmul(nxtseg->left.z - e->segment->left.z, old.x - nxtseg->left.x);
  c2 = fixmul(nxtseg->left.x - e->segment->left.x, e->y - nxtseg->left.z) - fixmul(nxtseg->left.z - e->segment->left.z, e->x - nxtseg->left.x);
  if (fixmul(c1, c2) <= 0) {
    e->x = old.x - ((e->x - old.x) << 1);
    e->y = old.y - ((e->y - old.y) << 1);
    if (e->life >= 2)
      e->life -= 2;
  }
  
  //Collision with right track edge
  c1 = fixmul(nxtseg->right.x - e->segment->right.x, old.y - nxtseg->right.z) - fixmul(nxtseg->right.z - e->segment->right.z, old.x - nxtseg->right.x);
  c2 = fixmul(nxtseg->right.x - e->segment->right.x, e->y - nxtseg->right.z) - fixmul(nxtseg->right.z - e->segment->right.z, e->x - nxtseg->right.x);
  if (fixmul(c1, c2) <= 0) {
    e->x = old.x - ((e->x - old.x) << 1);
    e->y = old.y - ((e->y - old.y) << 1);
    if (e->life >= 2)
      e->life -= 2;
  }

  //Moving into next segment
  c1 = fixmul(nxtseg->right.x - nxtseg->left.x, old.y - nxtseg->right.z) - fixmul(nxtseg->right.z - nxtseg->left.z, old.x - nxtseg->right.x);
  c2 = fixmul(nxtseg->right.x - nxtseg->left.x, e->y - nxtseg->right.z) - fixmul(nxtseg->right.z - nxtseg->left.z, e->x - nxtseg->right.x);
  //c3 = fixmul(old.x - e->x, nxtseg->left.z - old.y) - fixmul(old.y - e->y, nxtseg->left.x - old.x);
  //c4 = fixmul(old.x - e->x, nxtseg->right.z - old.y) - fixmul(old.y - e->y, nxtseg->right.x - old.x);
  if (fixmul(c1, c2) <= 0)
    e->segment = nxtseg;

  e->z += e->zz;
  e->r += e->rr; e->t += e->tt; e->p += e->pp;

/*  if (e->radius)
  {
    int x, y;
    rom_cell_t* cell;

#define IS_COLLISION ((e->z > cell->top) || (e->z < cell->bot))

    cell = &yeti->curr_map->cells[f2i(e->y)][f2i(x = e->x - e->radius)];

    if (IS_COLLISION)
    {
      e->x += FIXONE - (x & FIXCEIL);
      e->xx = 0;
    }

    cell = &yeti->curr_map->cells[f2i(e->y)][f2i(x = e->x + e->radius)];

    if (IS_COLLISION)
    {
      e->x -= (x & FIXCEIL);
      e->xx = 0;
    }

    cell = &yeti->curr_map->cells[f2i(y = e->y - e->radius)][f2i(e->x)];

    if (IS_COLLISION)
    {
      e->y += FIXONE - (y & FIXCEIL);
      e->yy = 0;
    }

    cell = &yeti->curr_map->cells[f2i(y = e->y + e->radius)][f2i(e->x)];

    if (IS_COLLISION)
    {
      e->y -= (y & FIXCEIL);
      e->yy = 0;
    }
  }*/
}

/******************************************************************************/

void yeti_clear_entities(yeti_t* yeti)
{
  int i;
  yeti->nentities = 0;
  yeti->root = 0;

  for (i = 0; i < YETI_ENTITY_MAX; i++)
    yeti->entities[i].parent = &yeti->entities[i];

  yeti->camera = yeti_entity(yeti,
    i2fdiv2(YETI_MAP_WIDTH),
    i2fdiv2(YETI_MAP_HEIGHT),
    i2f(0), 0);

  //for (i = 0; i < YETI_BULLET_MAX; i++)
  //{
  //  yeti->bullets[i] = yeti_entity(yeti, 0, 0, 0, 0);
  //}
}

void yeti_load_map(yeti_t* yeti, rom_map_t* src, h_map_t* hsrc)
{
  int x, y, i;

  yeti->curr_map = src;
  yeti->grid = hsrc;
  if (!yeti->camera) yeti_clear_entities(yeti);
  
  for (y = 0; y < YETI_MAP_HEIGHT; y++)
  {
    for (x = 0; x < YETI_MAP_WIDTH; x++)
    {
      if (yeti->curr_map->cells[y][x].ent == 1)
      {
        yeti->camera->x = i2f(x) + FIXHALF;
        yeti->camera->y = i2f(y) + FIXHALF;
      }
    }
  }
  yeti->camera->t = 0;
  yeti->camera->r = 0;
  yeti->camera->p = 0;

  for (i = 0; i < 100; i++)
    yeti->building_strength[i] = 100;
}

/*
** Name: yeti_init_map
** Desc: Clears a yeti map to its default settings.
*/
void yeti_init_map(yeti_t* yeti)
{
  int x, y;
  
  for (y = 0; y < YETI_MAP_HEIGHT; y++)
  {
    for (x = 0; x < YETI_MAP_WIDTH; x++)
    {
      cell_t* cell = &yeti->cells[y][x];
      cell_init(cell,
        x == 0 || y == 0 ||
        (x == YETI_MAP_WIDTH  - 1) ||
        (y == YETI_MAP_HEIGHT - 1));
    } 
  }
}

/******************************************************************************/

/*
** Name: isqrt
** Desc: Integer square root. Take the square root of an integer.
*/
int isqrt(int value)
{
  int root = 0;

#define STEP(shift) \
    if((0x40000000 >> shift) + root <= value)          \
    {                                                   \
        value -= (0x40000000 >> shift) + root;          \
        root = (root >> 1) | (0x40000000 >> shift);     \
    }                                                   \
    else                                                \
    {                                                   \
        root >>= 1;                                     \
    }

  STEP( 0); STEP( 2); STEP( 4); STEP( 6);
  STEP( 8); STEP(10); STEP(12); STEP(14);
  STEP(16); STEP(18); STEP(20); STEP(22);
  STEP(24); STEP(26); STEP(28); STEP(30);

  // round to the nearest integer, cuts max error in half

  if (root < value) root++;

  return root;
}

/******************************************************************************/

/*
** Name: yeti_init
** Desc: Setup a default map and position the camera.
*/
void yeti_init(
  yeti_t* yeti,
  framebuffer_t* front,
  framebuffer_t* back,
  texture_t* textures,
  palette_t palette,
  lua_t lua)
{
  int i;

  CLEARMEM(yeti);

  for (i = 0; i < YETI_TEXTURE_MAX; i++)
  {
    yeti->surfaces[i].xsize = 6;
    yeti->surfaces[i].ysize = 6;
  }

  yeti->surfaces[0].xsize = 4;
  yeti->surfaces[0].ysize = 4;
  yeti->surfaces[2].xsize = 5;
  yeti->surfaces[2].ysize = 5;

  //yeti_init_map(yeti);
  
  //yeti->viewport.front = front;
  //yeti->viewport.back  = back;

  yeti->textures    = textures;
  yeti->palette     = palette;
}

void yeti_tick(yeti_t* yeti)
{
  entity_t* e1;

  yeti->tick++;

  //traverse_binary_tree(yeti->root, entity_tick);
  //traverse_binary_tree(yeti->root, entity_motion);

  for (e1 = yeti->entities; e1 != &yeti->entities[YETI_ENTITY_MAX]; e1++)
    if (e1->parent != e1 && e1->ontick)
      e1->ontick(e1);

  for (e1 = yeti->entities; e1 != &yeti->entities[YETI_ENTITY_MAX]; e1++)
  {
    if (e1->parent == e1)
      continue;

  //for (e1 = yeti->head; e1 != 0; e1 = e1->next)
  //{
    entity_motion(e1);

    if (e1->radius)
    {
      entity_t* e2;
      
      for (e2 = e1+1; e2 != &yeti->entities[YETI_ENTITY_MAX]; e2++)
      {
        if (e2->radius)
        {
          int size = e1->radius + e2->radius;
          int x = e1->x - e2->x;
          int y = e1->y - e2->y;
          int z = e1->z - e2->z;

          if (
            x < size && x > -size &&
            y < size && y > -size &&
            z < size && z > -size )
          {
            if (e1->onhit) e1->onhit(e1, e2);
            if (e2->onhit) e2->onhit(e2, e1);

            /*if (!(e1->swi & ENTITY_SWI_NO_COLLISION_RESPONSE))
            {
              if (x > 0) {e1->x += ( size - x) >> 1;}
              if (x < 0) {e1->x += (-size - x) >> 1;}
              if (y > 0) {e1->y += ( size - y) >> 1;}
              if (y < 0) {e1->y += (-size - y) >> 1;}
            }
            if (!(e2->swi & ENTITY_SWI_NO_COLLISION_RESPONSE))
            {
              if (x > 0) {e2->x -= ( size - x) >> 1;}
              if (x < 0) {e2->x -= (-size - x) >> 1;}
              if (y > 0) {e2->y -= ( size - y) >> 1;}
              if (y < 0) {e2->y -= (-size - y) >> 1;}
            }*/
          }
        }
      }
    }
  }
  if (yeti->indoors_model == 0)
    camera_behaviour(yeti->camera);
  yeti->keyboard._pad = yeti->keyboard.pad;
}

/******************************************************************************/

void* yeti_memcpy(void* d, void* s, unsigned n)
{
  while (n--) ((u8*)d)[n] = ((u8*)s)[n];
  return d;
}

void* yeti_memset(void *s, int c, unsigned n)
{
  while (n--) ((u8*)s)[n] = c;
  return s;
}

/******************************************************************************/
